@@ -3,6 +3,7 @@ import hashlib
 import logging
 import pprint
 import random
+import re
 import time
 
 from django.db import DEFAULT_DB_ALIAS, DatabaseError
@@ -59,14 +60,59 @@ class CursorWrapper:
                 self._cursor = self._backend.create_secondary_cursor()
         return self._cursor
 
-    def add_history(self, item):
-        if self._backend.failover_active:
+    def add_history(self, attr, args, kwargs, return_value):
+        if self._backend.failover_active and attr != '_executed':
             if self._history_entry_index is None or len(self._backend.failover_history) == 0:
                 self._history_entry_index = len(self._backend.failover_history)
                 self._backend.failover_history.append([])
-            self._backend.failover_history[self._history_entry_index].append(item)
+            self._backend.failover_history[self._history_entry_index].append((
+                attr,
+                args,
+                kwargs,
+                hashlib.sha1(pprint.pformat(return_value).encode()).hexdigest()
+            ))
             self._backend.failover_history_size += 1
 
+            # store the insert id of an auto field, so it does not change when the history is replayed
+            if self._backend.failover_history[self._history_entry_index][-1][0] in ('fetchone', 'fetchall') \
+                    and self._backend.failover_history[self._history_entry_index][-2][0] == 'execute' \
+                    and self._backend.failover_history[self._history_entry_index][-2][1][0].startswith('INSERT '):
+                insert_entry = self._backend.failover_history[self._history_entry_index][-2]
+                insert_sql = insert_entry[1][0]
+                match = re.match(
+                    r'^INSERT INTO `([^`]+)` \((.+)\) VALUES (.+) RETURNING `([^`]+)`.`([^`]+)`$',
+                    insert_sql
+                )
+                if match:
+                    table_name, fields, values, auto_table, auto_field = match.groups()
+                    values_new = values.replace('(%s, ', '(%s, %s, ')
+                    new_sql = (
+                        f'INSERT INTO `{table_name}` '
+                        f'(`{auto_field}`, {fields}) '
+                        f'VALUES {values_new} '
+                        f'RETURNING `{auto_table}`.`{auto_field}`'
+                    )
+                    if new_sql != insert_sql:
+                        kwargs = copy.deepcopy(insert_entry[2])
+                        kwargs['args'] = list(kwargs['args'])
+                        values_count = int(round(len(kwargs['args']) / len(return_value)))
+                        if len(return_value) == 1:
+                            kwargs['args'].insert(0, str(return_value[0]))
+                        else:
+                            for x in range(len(return_value)):
+                                kwargs['args'].insert(x * values_count + x, str(return_value[x][0]))
+                        kwargs['args'] = tuple(kwargs['args'])
+                        self._backend.failover_history[self._history_entry_index][-2] = (
+                            insert_entry[0],
+                            (new_sql,),
+                            kwargs,
+                            insert_entry[3],
+                        )
+                    else:
+                        LOGGER.warning('SQL unchanged: %s' % insert_sql)
+                else:
+                    LOGGER.warning('No match: %s' % insert_sql)
+
     def prepare(self, query):
         if query is None:
             rw_query = True
@@ -122,24 +168,13 @@ class CursorWrapper:
                 except Exception as exc:
                     self._handle_exc(exc)
                     ret = getattr(self._cursor, func.__name__)(*args, **kwargs)
-                self.add_history((
-                    item,
-                    args,
-                    kwargs,
-                    hashlib.sha1(pprint.pformat(ret).encode()).hexdigest()
-                ))
+                self.add_history(item, args, kwargs, ret)
                 return ret
             return decor
         if hasattr(obj, '__call__'):
             obj = wrap(obj)
         else:
-            self.add_history((
-                item,
-                None,
-                None,
-                hashlib.sha1(pprint.pformat(obj).encode()).hexdigest()
-            ))
-
+            self.add_history(item, None, None, obj)
         return obj
 
     def __enter__(self):
@@ -163,17 +198,19 @@ class CursorWrapper:
         if self._in_handle_exc or not exc.args:
             raise exc
         self._in_handle_exc = True
-        if self._backend.failover_active and len(exc.args) and len(self._backend.failover_history):
+        if self._backend.failover_active and len(exc.args):
             error_code = str(exc.args[0])
             if error_code in (
                     '1047',  # Unknown command (wsrep_reject_queries)
-                    # '1205',  # Lock wait timeout exceeded; try restarting transaction
-                    # '1213',  # Deadlock found when trying to get lock; try restarting transaction
+                    '1180',  # Got error 6 "No such device or address" during COMMIT
+                    '1205',  # Lock wait timeout exceeded; try restarting transaction
+                    '1213',  # Deadlock found when trying to get lock; try restarting transaction
                     '2006',  # MySQL server has gone away
                     '2013',  # Lost connection to MySQL server during query
             ):
                 autocommit = self._backend.autocommit
                 in_atomic_block = self._backend.in_atomic_block
+                in_write_transaction = self._backend.in_write_transaction
                 needs_rollback = self._backend.needs_rollback
                 savepoint_ids = self._backend.savepoint_ids
                 history = copy.deepcopy(self._backend.failover_history)
@@ -204,6 +241,7 @@ class CursorWrapper:
                     self._backend.connection.commit()
                     self._backend.set_autocommit(autocommit)
                 self._backend.in_atomic_block = in_atomic_block
+                self._backend.in_write_transaction = in_write_transaction
                 self._backend.savepoint_ids = savepoint_ids
                 self._backend.failover_history = copy.deepcopy(history)
                 self._backend.failover_history_size = history_size
@@ -364,40 +402,35 @@ class DatabaseWrapper(base.DatabaseWrapper):
         return self._secondary_wrapper
 
     def _set_autocommit(self, autocommit):
+        super(DatabaseWrapper, self)._set_autocommit(autocommit)
         if autocommit:
-            self.sync_wait_primary()
             self.in_write_transaction = False
         if not autocommit and not self.optimistic_transactions:
             self.in_write_transaction = True
         if autocommit != self.autocommit:
             self.failover_history_reset()
             self.failover_active = self.failover_enable
-        return super(DatabaseWrapper, self)._set_autocommit(autocommit)
-
-    def sync_wait_primary(self):
-        if self.wsrep_sync_after_write and not self.primary_synced:
-            try:
-                t = time.perf_counter()
-                self._wsrep_sync_wait(self.connection)
-                t = time.perf_counter() - t
-                LOGGER.debug('Primary synced in %f seconds' % t)
-            except Exception as e:
-                LOGGER.warning('Error while syncing primary: %s' % str(e), exc_info=True)
-            self.primary_synced = True
 
     def sync_wait_secondary(self):
         if self.wsrep_sync_after_write and not self.secondary_synced:
-            try:
-                t = time.perf_counter()
-                if self.wsrep_sync_use_gtid:
-                    self._wsrep_sync_wait_upto_gtid()
-                else:
-                    self._wsrep_sync_wait(self.secondary_wrapper.connection)
-                t = time.perf_counter() - t
-                LOGGER.debug('Secondary synced in %f seconds' % t)
-            except Exception as e:
-                LOGGER.warning('Error while syncing secondary: %s' % str(e), exc_info=True)
-            self.secondary_synced = True
+            t = time.perf_counter()
+            retry = 0
+            while not self.secondary_synced:
+                try:
+                    if self.wsrep_sync_use_gtid:
+                        self._wsrep_sync_wait_upto_gtid()
+                    else:
+                        self._wsrep_sync_wait(self.secondary_wrapper.connection)
+                    self.secondary_synced = True
+                except Exception as e:
+                    error_code = str(e.args[0]) if e.args else ''
+                    if retry < 3 and error_code == '1205':  # Lock wait timeout exceeded; try restarting transaction
+                        LOGGER.info('Retry syncing secondary after: %s' % str(e), exc_info=True)
+                        retry += 1
+                        continue
+                    raise e
+            t = time.perf_counter() - t
+            LOGGER.debug('Secondary synced in %f seconds' % t)
 
     def replay_history(self, history):
         for x, entry in enumerate(history, start=1):
@@ -417,11 +450,14 @@ class DatabaseWrapper(base.DatabaseWrapper):
                 cursor.close()
             else:
                 return cursor
+        # return a new cursor if history is empty
+        return self.connection.cursor()
 
     def _wsrep_sync_wait(self, connection):
         with connection.cursor() as cursor:
             cursor.execute(
                 'SET @wsrep_sync_wait_orig = @@wsrep_sync_wait;'
+                'SET SESSION lock_wait_timeout = 5;'
                 'SET SESSION wsrep_sync_wait = GREATEST(@wsrep_sync_wait_orig, 1);'
                 'SELECT 1;'
                 'SET SESSION wsrep_sync_wait = @wsrep_sync_wait_orig;'
@@ -429,7 +465,7 @@ class DatabaseWrapper(base.DatabaseWrapper):
 
     def _wsrep_sync_wait_upto_gtid(self):
         with self.connection.cursor() as primary_cursor:
-            primary_cursor.execute('SELECT WSREP_LAST_WRITTEN_GTID()')
+            primary_cursor.execute('SELECT WSREP_LAST_SEEN_GTID()')
             result = primary_cursor.fetchone()
             primary_gtid = result[0].decode('utf-8')
         with self.secondary_wrapper.connection.cursor() as secondary_cursor:
